#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

/************************************************************/

size_t cs_hash(char *str) {
  size_t result = 0;
  while (*str) {
    result += *str;
    str++;
  }
  return result;
}

int cs_is_equivalent(char *str1, char *str2) {
  while (*str1 != 0) {
    if (*str1 != *str2)
      return 0;
    str1++;
    str2++;
  }
  return *str2 == 0;
}

size_t ci_hash(char *str) {
  size_t result = 0;
  while (*str != 0) {
    result += tolower((unsigned char)*str);
    str++;
  }
  return result;
}

int ci_is_equivalent(char *str1, char *str2) {
  while (*str1 != 0) {
    if (tolower((unsigned char)*str1) != tolower((unsigned char)*str2))
      return 0;
    str1++;
    str2++;
  }
  return *str2 == 0;
}

size_t hash2(char *str) {
  int result = 0;
  while (*str != 0) {
    result++;
    str++;
  }
  return result | 0x1;
}

/************************************************************/

typedef size_t (*get_hash_t)(char *str);
typedef size_t (*get_hash2_t)(char *str); /* must return an odd number */
typedef int (*is_equivalent_t)(char *str1, char *str2);

typedef struct {
  size_t set_size; /* always < array_size * FILL_FACTOR */
  size_t array_size; /* always a power of two */
  char **strs;
  get_hash_t get_hash;
  get_hash2_t get_hash2;
  is_equivalent_t is_equivalent;
} hash_set_t;

#define FILL_FACTOR 0.8

hash_set_t *make_hash_set(get_hash_t get_hash,
                          get_hash2_t get_hash2,
                          is_equivalent_t is_equivalent)
{
  hash_set_t *hs = malloc(sizeof(hash_set_t));

  hs->set_size = 0;
  hs->array_size = 16;
  hs->strs = calloc(hs->array_size, sizeof(char *));
  hs->get_hash = get_hash;
  hs->get_hash2 = get_hash2;
  hs->is_equivalent = is_equivalent;

  return hs;
}

void hash_set_clear(hash_set_t *hs)
{
  hs->set_size = 0;
  memset(hs->strs, 0, hs->array_size * sizeof(char*));
}

void grow_and_rehash(hash_set_t *hs);

int hash_set_add(hash_set_t *hs, char *str) {
  size_t i;

  i = hs->get_hash(str);
  while (hs->strs[i & (hs->array_size-1)]
         && !hs->is_equivalent(str, hs->strs[i & (hs->array_size-1)])) {
    i += hs->get_hash2(str);
  }

  if (!hs->strs[i & (hs->array_size-1)]) {
    if ((hs->set_size + 1) >= (FILL_FACTOR * hs->array_size)) {
      grow_and_rehash(hs);
      hash_set_add(hs, str);
    } else {
      hs->set_size++;
      hs->strs[i & (hs->array_size-1)] = str;
    }
    return 1;
  } else
    return 0;
}

void grow_and_rehash(hash_set_t *hs)
{
  size_t old_array_size = hs->array_size, i;
  char **old_strs = hs->strs;

  hs->set_size = 0;
  hs->array_size *= 2;
  hs->strs = calloc(hs->array_size, sizeof(char*));

  for (i = 0; i < old_array_size; i++) {
    if (old_strs[i])
      hash_set_add(hs, old_strs[i]);
  }

  free(old_strs);
}

/************************************************************/

#include "bm_strings.h"

int main() {
  hash_set_t *cs_hs, *ci_hs;
  int i, j;

  cs_hs = make_hash_set(cs_hash, hash2, cs_is_equivalent);
  ci_hs = make_hash_set(ci_hash, hash2, ci_is_equivalent);

  for (j = 0; j < ITERS; j++) {
    hash_set_clear(cs_hs);
    hash_set_clear(ci_hs);
    for (i = 0; bm_strings[i] != NULL; i++) {
      hash_set_add(cs_hs, bm_strings[i]);
      hash_set_add(ci_hs, bm_strings[i]);
    }
  }

  printf("%ld vs. %ld\n", cs_hs->set_size, ci_hs->set_size);

  return 0;
}
